/*----------------------------------------------------------------------
  File    : xsom.c
  Contents: self-organizing map visualization program
  Author  : Christian Borgelt
  History : 2000.06.17 file created from file xlvq.c
            2000.06.18 load routine for self-organizing maps added
            2001.07.08 sphere surface added as a training pattern region
            2001.07.09 switching between pattern regions improved
            2001.07.13 scaling factor for mouse movement introduced
            2001.12.21 cube region added
            2002.10.22 menu item for METAPOST output added
            2013.08.09 adapted to higher compiler warning level
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "random.h"
#ifndef SOM_DRAW
#define SOM_DRAW
#endif
#ifndef SOM_MPOST
#define SOM_MPOST
#endif
#include "somd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI      3.14159265358979323846  /* \pi */
#endif

/* --- extensions --- */
#define MINXEXT      120        /* minimal x-extension of window */
#define MINYEXT       90        /* minimal y-extension of window */
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     320        /* initial y-extension of window */

/* --- movement modes --- */
#define ROTATE         1        /* change heading and pitch */
#define YAW            2        /* change heading */
#define PITCH          3        /* change pitch angle */
#define ROLL           4        /* change roll  angle */
#define ZOOM           5        /* change distance to the origin */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_WIDGET     (-6)       /* widget creation failed */
#define E_DIALOG     (-7)       /* cannot create dialog box */
#define E_COLOR      (-8)       /* color allocation failed */
#define E_GRAPH      (-9)       /* cannot create graphics context */
#define E_SAVE      (-10)       /* no self-organizing map to save */
#define E_UNKNOWN   (-11)       /* unkown error */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Action Function Prototypes
----------------------------------------------------------------------*/
static void mvmode (Widget w, XEvent *e, String *s, Cardinal *c);
static void mouse  (Widget w, XEvent *e, String *s, Cardinal *c);
static void load   (Widget w, XEvent *e, String *s, Cardinal *c);
static void save   (Widget w, XEvent *e, String *s, Cardinal *c);
static void mpost  (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit   (Widget w, XEvent *e, String *s, Cardinal *c);
static void init   (Widget w, XEvent *e, String *s, Cardinal *c);
static void exec   (Widget w, XEvent *e, String *s, Cardinal *c);
static void view   (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw (Widget w, XEvent *e, String *s, Cardinal *c);
static void grid   (Widget w, XEvent *e, String *s, Cardinal *c);
static void params (Widget w, XEvent *e, String *s, Cardinal *c);
static void about  (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xsom.rsc"             /* fallback resources */

static XtActionsRec actions[] = {
  { "mode",     mvmode   },     /* set movement mode */
  { "mouse",    mouse    },     /* mouse movement */
  { "load",     load     },     /* File     > Load Map... */
  { "save",     save     },     /* File     > Save Map... */
  { "mpost",    mpost    },     /* File     > Save as Metapost... */
  { "quit",     quit     },     /* File     > Quit */
  { "init",     init     },     /* Actions  > Initialize */
  { "view",     view     },     /* Actions  > Reset View */
  { "redraw",   redraw   },     /* Actions  > Redraw */
  { "exec",     exec     },     /* Actions  > Start/Stop Training */
  { "grid",     grid     },     /* Settings > Grid/Shape... */
  { "params",   params   },     /* Settings > Parameters... */
  { "about",    about    },     /* Help     > About XSOM... */
  { "db_focus", db_focus },     /* set focus to input gadget */
  { "db_next",  db_next  },     /* switch to next input gadget */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "XSOM.Error",
                         "no error",                               0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XSOM.Error",
                         "not enough memory",                      0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XSOM.Warning",
                         "cannot open file:\n%s",                  1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XSOM.Warning",
                         "read error on file:\n%s",                1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XSOM.Warning",
                         "write error on file:\n%s",               1 },
  /* E_INIT     -5 */  { "error",   "init",    "XSOM.Error",
                         "initialization failed",                  0 },
  /* E_WIDGET   -6 */  { "error",   "widget",  "XSOM.Error",
                         "widget creation failed",                 0 },
  /* E_DIALOG   -7 */  { "warning", "dialog",  "XSOM.Warning",
                         "cannot create dialog box",               0 },
  /* E_COLOR    -8 */  { "warning", "color",   "XSOM.Warning",
                         "color allocation failed",                0 },
  /* E_GRAPH    -9 */  { "error",   "gc",      "XSOM.Error",
                         "cannot create graphics context",         0 },
  /* E_SAVE    -10 */  { "warning", "som",     "XSOM.Warning",
                         "no self-organizing map created yet",     0 },
  /* E_UNKNOWN -11 */  { "error", "unknown",   "XSOM.Error",
                         "unknown error",                          0 }};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for drawing */

/* --- property sheets --- */
static PSHEET  *psh_grid   = NULL;  /* grid and shape */
static PSHEET  *psh_params = NULL;  /* learning parameters */

/* --- color information --- */
static unsigned long black;     /* black and white */
static unsigned long white;     /* for normal drawing */
static unsigned long grey;      /* grey for background shape */

/* --- neural network information --- */
static struct {                 /* -- parameters -- */
  int    shape;                 /* shape of pattern region */
  int    width, height;         /* width and height of the map */
  double range;                 /* initial weight range */
  double lrate;                 /* initial learning rate */
  double lrdec;                 /* learning rate decay parameter */
  int    nbhood;                /* neighborhood distance function */
  double radius;                /* initial neighborhood radius */
  double radec;                 /* radius decay parameter */
  int    patcnt;                /* patterns between redraws */
  double delay;                 /* delay (in seconds) */
} nnpar = { SOM_SQUARE, 10, 10, 0.5, 0.6, -0.1,
            SOM_GAUSSIAN, 2.5, -0.1, 1, 0.05 };

static SOM    *som    = NULL;   /* self-organizing map */
static int    patcnt  = 0;      /* number of processed patterns */
static int    running = 0;      /* whether training is running */
static Pixmap buffer  = None;   /* buffer for drawing */
static Drawable drawable;       /* either window or buffer */
static XtIntervalId iid;        /* id of time interval procedure */
static PROJ3D *proj3d = NULL;   /* 3D to 2D projection */
static int    mode    = 0;      /* movement mode */
static int    mx = 0, my = 0;   /* old mouse coordinates */
static double xoff, yoff;       /* coordinate offsets */
static double xscl, yscl;       /* scaling factors */

/*----------------------------------------------------------------------
  Error Reporting Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (!XtIsTransientShell(widget))
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close the dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable the menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data */
  if (!msg->name) msg = errmsgs -E_UNKNOWN;
  va_start(va, code);           /* get list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Training Function
----------------------------------------------------------------------*/

static void train (XtPointer data, XtIntervalId *id)
{                               /* --- do one training step */
  int    i;                     /* loop variable */
  double x, y, z;               /* coordinates of training pattern */
  double lrate;                 /* current learning rate */
  double radius;                /* current neighborhood radius */

  if (!running || !som)         /* check whether training is running */
    return;                     /* and whether there is a s.o. map */
  for (i = nnpar.patcnt; --i >= 0; ) {
    som_genpat(nnpar.shape, drand, &x, &y, &z);
    patcnt++;                   /* count the pattern */
    lrate  = (nnpar.lrdec <= 0) /* compute the learning rate */
           ? nnpar.lrate  *pow(patcnt, nnpar.lrdec)
           : nnpar.lrate  *pow(nnpar.lrdec, patcnt);
    radius = (nnpar.radec <= 0) /* compute the neighborhood radius */
           ? nnpar.radius *pow(patcnt, nnpar.radec)
           : nnpar.radius *pow(nnpar.radec, patcnt);
    som_train(som, x, y, z, lrate, nnpar.nbhood, radius);
  }                             /* train the self-organizing map */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
  iid = XtAppAddTimeOut(appctx, (unsigned long)(nnpar.delay *1000),
          train, NULL);         /* reregister the interval procedure */
}  /* train() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Map...' callback */
  FILE *file;                   /* output file */
  SOM  *tmp;                    /* temp. buffer for self-org. map */
  int  shape;                   /* id of shape of pattern region */

  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  file = fopen(fname, "r");     /* open the input file */
  if (!file) { error(E_FOPEN, fname); return; }
  if ((fscanf(file, "%i", &shape) != 1)
  ||  (shape < SOM_SQUARE)      /* read the id of the shape */
  ||  (shape > SOM_CUBE)) {     /* of the pattern region */
    error(E_FREAD, fname); return; }
  tmp = som_load(file);         /* load the self-org. map */
  fclose(file);                 /* close the input file */
  if (!tmp) { error(E_FREAD, fname); return; }
  if (som) som_delete(som);     /* delete an existing self-org. map */
  som         = tmp;            /* and set the loaded one */
  nnpar.shape = shape;          /* set shape of the pattern region */
  patcnt      = 0;              /* clear the pattern counter and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save Map...' callback */
  FILE *file;                   /* output file */

  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  file = fopen(fname, "w");     /* open the output file */
  if (!file) { error(E_FOPEN, fname); return; }
  fprintf(file, "%d ", nnpar.shape);
  som_save(som, file);          /* save id of shape of pattern region */
  if (fclose(file) != 0)        /* and the self-organizing map, */
    error(E_FWRITE);            /* then close the output file */
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_mpost (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save as Metapost...' callback */
  FILE *file;                   /* output file */

  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  file = fopen(fname, "w");     /* open the output file */
  if (!file) { error(E_FOPEN, fname); return; }
  som_mpost(som, (nnpar.shape >= SOM_SPHERE) ? proj3d : NULL, file);
  if (fclose(file) != 0)        /* save the self-organizing map, */
    error(E_FWRITE);            /* then close the output file */
}  /* dcb_mpost() */

/*--------------------------------------------------------------------*/

static void dcb_grid (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Grid/Shape...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  switch ((ptrdiff_t)(psh_getval(psh_grid, "shape"))) {
    case  1: nnpar.shape = SOM_TRIANGLE; break;
    case  2: nnpar.shape = SOM_CIRCLE;   break;
    case  3: nnpar.shape = SOM_SPHERE;   break;
    case  4: nnpar.shape = SOM_PARABOLA; break;
    case  5: nnpar.shape = SOM_CUBIC;    break;
    case  6: nnpar.shape = SOM_COSINE;   break;
    case  7: nnpar.shape = SOM_CUBE;     break;
    default: nnpar.shape = SOM_SQUARE;   break;
  }                             /* shape of the pattern region */
  nnpar.width  = atoi(psh_getval(psh_grid, "width"));
  if      (nnpar.width  < 1)   nnpar.width  = 1;
  else if (nnpar.width  > 100) nnpar.width  = 100;
                                /* width  of self-org. map (in units) */
  nnpar.height = atoi(psh_getval(psh_grid, "height"));
  if      (nnpar.height < 1)   nnpar.height = 1;
  else if (nnpar.height > 100) nnpar.height = 100;
                                /* height of self-org. map (in units) */
  if (som && ((som_width(som)  != nnpar.width)
  ||          (som_height(som) != nnpar.height))) {
    som_delete(som); som = NULL; }  /* delete the self-organizing map */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_grid() */

/*--------------------------------------------------------------------*/

static void dcb_params (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Parameters...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  nnpar.range  = atof(psh_getval(psh_params, "range"));
  if      (nnpar.range < 0)     nnpar.range  = 0;
  else if (nnpar.range > 100)   nnpar.range  = 100;
                                /* initial weight range */
  nnpar.lrate  = atof(psh_getval(psh_params, "lrate"));
  if      (nnpar.lrate < 1e-12) nnpar.lrate  = 1e-12;
  else if (nnpar.lrate > 10)    nnpar.lrate  = 10;
                                /* inital learning rate */
  nnpar.lrdec  = atof(psh_getval(psh_params, "lrdec"));
  if      (nnpar.lrdec > 1)     nnpar.lrdec  = 1;
                                /* learning rate decay parameter */
  nnpar.nbhood = ((ptrdiff_t)psh_getval(psh_params, "nbhood") == 0)
               ? SOM_RECT : SOM_GAUSSIAN;
                                /* neighborhood type */
  nnpar.radius = atof(psh_getval(psh_params, "radius"));
  if      (nnpar.radius < 0)    nnpar.radius = 0;
  else if (nnpar.radius > 100)  nnpar.radius = 100;
                                /* initial neighborhood radius */
  nnpar.radec  = atof(psh_getval(psh_params, "radec"));
  if      (nnpar.radec  > 1)    nnpar.radec  = 1;
                                /* radius decay parameter */
  nnpar.patcnt = atoi(psh_getval(psh_params, "patcnt"));
  if      (nnpar.patcnt < 1)    nnpar.patcnt = 1;
  else if (nnpar.patcnt > 1000) nnpar.patcnt = 1000;
                                /* patterns between redraws */
  nnpar.delay = atof(psh_getval(psh_params, "delay"));
  if      (nnpar.delay < 0.001) nnpar.delay  = 0.001;
  else if (nnpar.delay > 10)    nnpar.delay  = 10;
                                /* delay between redraws */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_params() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ load(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ save(NULL, NULL, NULL, NULL); }

static void mcb_mpost  (Widget widget, XtPointer client, XtPointer call)
{ mpost(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_init   (Widget widget, XtPointer client, XtPointer call)
{ init(NULL, NULL, NULL, NULL); }

static void mcb_start  (Widget widget, XtPointer client, XtPointer call)
{ exec(NULL, NULL, NULL, NULL); }

static void mcb_view   (Widget widget, XtPointer client, XtPointer call)
{ view(NULL, (XEvent*)1, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_grid   (Widget widget, XtPointer client, XtPointer call)
{ grid(NULL, NULL, NULL, NULL); }

static void mcb_params (Widget widget, XtPointer client, XtPointer call)
{ params(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void mvmode (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- set movement mode */
  if (!c || (*c != 1)) return;  /* check for an argument */
  if      (strcmp(*s, "rotate") == 0) mode = ROTATE;
  else if (strcmp(*s, "yaw")    == 0) mode = YAW;
  else if (strcmp(*s, "pitch")  == 0) mode = PITCH;
  else if (strcmp(*s, "roll")   == 0) mode = ROLL;
  else if (strcmp(*s, "zoom")   == 0) mode = ZOOM;
  else                                mode = 0;
  mx = e->xbutton.x;            /* note the movement mode and */
  my = e->xbutton.y;            /* the current mouse position */
  p3d_save(proj3d);             /* and save the current view */
}  /* mvmode() */

/*--------------------------------------------------------------------*/

static void mouse (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- evaluate mouse movement */
  int    dx, dy;                /* pixels mouse was moved */
  double sf;                    /* scaling factor */

  if (mode <= 0) return;        /* check for a movement mode */
  dx = e->xbutton.x -mx;        /* get the number of pixels */
  dy = e->xbutton.y -my;        /* the mouse was moved */
  sf = 64/xscl;                 /* compute the scaling factor */
  switch (mode) {               /* according to the movement mode */
    case ROTATE: if ((dx == 0) && (dy == 0)) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case YAW:    if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case PITCH:  if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH); break;
    case ROLL:   if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_roll (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case ZOOM:   if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_move (proj3d, 0, dy*0.02, 0);              break;
  }                             /* change the viewing parameters */
  redraw(NULL,NULL,NULL,NULL);  /* and redraw the self-organizing map */
}  /* mouse() */

/*--------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Map... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* load() */                 /* open a file select dialog */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save Map... */
  const char *t;                /* buffer for dialog box title */

  if (!som) {                   /* if there is no map, */
    error(E_SAVE); return; }    /* abort the function */
  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* save() */                 /* open a file select dialog */

/*--------------------------------------------------------------------*/

static void mpost (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save Map... */
  const char *t;                /* buffer for dialog box title */

  if (!som) {                   /* if there is no map, */
    error(E_SAVE); return; }    /* abort the function */
  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*mpost"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_mpost) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* mpost() */                /* open a file select dialog */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  if (som) som_delete(som);     /* delete a self-organizing map */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate the program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void init (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Initialize */
  if (!som) {                   /* if there is no self-org. map */
    som = som_create(nnpar.width, nnpar.height);
    if (!som) { error(E_NOMEM); return; }
  }                             /* create a self-organizing map */
  som_init(som, drand, nnpar.range, nnpar.range,
           (nnpar.shape >= SOM_SPHERE) ? nnpar.range : 0);
  patcnt = 0;                   /* clear the pattern counter and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* init() */

/*--------------------------------------------------------------------*/

static void exec (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Start/Stop Training */
  if (running) {                /* if running, remove procedure */
    mn_visible(menu, MN_ITEMS, 3);
    mn_enable (menu, MN_ITEMS, 3);
    running = 0; XtRemoveTimeOut(iid); }
  else {                        /* otherwise start training */
    mn_visible(menu, MN_ITEMS, 5);
    mn_enable (menu, MN_ITEMS, 4);
    if (!som) init(NULL, NULL, NULL, NULL);
    if ( som) { running = 1; train(NULL, NULL); }
  }
}  /* exec() */

/*--------------------------------------------------------------------*/

static void view (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Reset View */
  p3d_eye  (proj3d, 0, 0, 5);   /* initialize eye, view, and frustrum */
  p3d_view (proj3d, 0, -M_PI/2, 0);
  p3d_frust(proj3d, M_PI/6, 1e-6, 1e6);
  if (e) redraw(NULL,NULL,NULL,NULL);
}  /* view() */                 /* redraw the window contents */

/*--------------------------------------------------------------------*/

static void drawpt (double x, double y, int line)
{                               /* --- move to/draw a line to a point */
  static int px = 0, py = 0;    /* coordinates of preceding point */
  int        cx, cy;            /* coordinates of current point */

  cx = (int)(xoff +x *xscl);    /* transform the coordinates */
  cy = (int)(yoff -y *((nnpar.shape >= SOM_SPHERE) ? xscl : yscl));
  if (line)                     /* if to draw a line to new point */
    XDrawLine(display, drawable, gc, px, py, cx, cy);
  px = cx; py = cy;             /* note coordinates of current point */
}  /* drawpt() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static Dimension owd = 0;     /* old window width */
  static Dimension oht = 0;     /* old window height */
  XEvent     de;                /* dummy event */
  XRectangle rc;                /* clipping rectangle */
  Dimension  wd, ht;            /* window extensions */
  int        x0, x1, xm;        /* horizontal coordinates */
  int        y0, y1;            /* vertical coordinates */

  /* --- initialize graphics --- */
  if (!gc) {                    /* if there is no graphics context */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  XtVaGetValues(w_view, XtNwidth, &wd, XtNheight, &ht, NULL);
  if (e && (e->type == Expose)  /* if the window was exposed, */
  &&  (wd == owd) && (ht == oht)) {        /* but not resized */
    rc.x      = (short)e->xexpose.x;       /* build and set a */
    rc.y      = (short)e->xexpose.y;    /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted); }
  else {                        /* if the window was resized */
    rc.x = rc.y = 0;            /* or this is a forced redraw, */
    rc.width  = wd;             /* redraw all of the window */
    rc.height = ht;             /* (set full window rectangle) */
    while (XCheckTypedWindowEvent(display, window, Expose, &de));
  }                             /* remove all exposure events */
  if ((wd != owd) || (ht != oht)/* if the window extensions changed, */
  ||  (buffer == None)) {       /* a new buffer has to be allocated */
    if (buffer != None) XFreePixmap(display, buffer);
    x0       = XDefaultDepth(display, screen);
    buffer   = XCreatePixmap(display, window, wd, ht, (unsigned)x0);
    drawable = (buffer == None) ? window : buffer;
  }                             /* if no buffer can be allocated, */
                                /* draw directly into the window */
  owd  = wd; oht = ht;          /* note the (new) window dimensions */
  xoff = 0.5*wd; yoff = 0.5*ht;
  xscl = 0.476*(wd-1);          /* compute the coordinate offsets */
  yscl = 0.476*(ht-1);          /* and the scaling factors */

  /* --- draw background --- */
  XSetForeground(display, gc, white);
  XFillRectangle(display, drawable, gc, 0, 0, wd, ht);

  /* --- draw pattern region --- */
  XSetForeground(display, gc, grey);
  x0 = (int)(xoff -xscl); x1 = (int)(xoff +xscl);
  y0 = (int)(yoff +yscl); y1 = (int)(yoff -yscl);
  xm = (int)(xoff);             /* compute coordinates */
  switch (nnpar.shape) {        /* evaluate shape */
    case SOM_SQUARE:            /* square region */
      XDrawLine(display, drawable, gc, x0, y0, x1, y0);
      XDrawLine(display, drawable, gc, x1, y0, x1, y1);
      XDrawLine(display, drawable, gc, x1, y1, x0, y1);
      XDrawLine(display, drawable, gc, x0, y1, x0, y0); break;
    case SOM_TRIANGLE:          /* triangular region */
      XDrawLine(display, drawable, gc, x0, y0, x1, y0);
      XDrawLine(display, drawable, gc, x0, y0, xm, y1);
      XDrawLine(display, drawable, gc, x1, y0, xm, y1); break;
    case SOM_CIRCLE:            /* circular region */
      XDrawArc (display, drawable, gc, x0, y1,
                (unsigned)(x1-x0+1), (unsigned)(y0-y1+1),
                0, 360*64);                             break;
  }

  /* --- draw self-organizing map --- */
  if (som) {                    /* if there is a self-org. map */
    XSetForeground(display, gc, black);
    som_draw(som, (nnpar.shape >= SOM_SPHERE) ? proj3d : NULL, drawpt);
  }                             /* draw the self-organizing map */
  if (drawable != window)       /* copy buffer contents to window */
    XCopyArea(display, drawable, window, gc, 0, 0, wd, ht, 0, 0);

  /* --- print the number of patterns --- */
  sprintf(buf, "%d", patcnt);   /* set the menu info entry */
  mn_info(menu, buf); mn_resize(menu, wd);

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);  /* and clear the clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void grid (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Grid/Shape... */
  char buf[32];                 /* temporary buffer */

  if (!psh_grid) {              /* if property sheet not created yet */
    psh_grid = psh_create("gridDlg", w_top, 1);
    if (!psh_grid) return;      /* create a property sheet */
    if ((psh_addline(psh_grid, "lshape")              != 0)
    ||  (psh_additem(psh_grid, "shape",  PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "square")    != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "triangle")  != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "circle")    != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "sphere")    != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "parabola")  != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "cubic")     != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "cosine")    != 0)
    ||  (psh_menuadd(psh_grid, "shape",  "cube")      != 0)
    ||  (psh_addline(psh_grid, "lwidth")              != 0)
    ||  (psh_additem(psh_grid, "width",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_grid, "lheight")             != 0)
    ||  (psh_additem(psh_grid, "height", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_grid); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  switch (nnpar.shape) {        /* shape of the pattern region */
    case SOM_TRIANGLE: psh_setval(psh_grid, "shape", (char*)1); break;
    case SOM_CIRCLE  : psh_setval(psh_grid, "shape", (char*)2); break;
    case SOM_SPHERE  : psh_setval(psh_grid, "shape", (char*)3); break;
    case SOM_PARABOLA: psh_setval(psh_grid, "shape", (char*)4); break;
    case SOM_CUBIC   : psh_setval(psh_grid, "shape", (char*)5); break;
    case SOM_COSINE  : psh_setval(psh_grid, "shape", (char*)6); break;
    case SOM_CUBE    : psh_setval(psh_grid, "shape", (char*)7); break;
    default          : psh_setval(psh_grid, "shape", (char*)0); break;
  }
  sprintf(buf, "%d", nnpar.width);   /* map width  (in units) */
  psh_setval(psh_grid, "width",  buf);
  sprintf(buf, "%d", nnpar.height);  /* map height (in units) */
  psh_setval(psh_grid, "height", buf);
  psh_handle(psh_grid, dcb_grid);    /* handle the dialog */
}  /* grid() */

/*--------------------------------------------------------------------*/

static void params (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Parameters... */
  char buf[32];                 /* buffer for numbers */

  if (!psh_params) {            /* if property sheet not created yet */
    psh_params = psh_create("paramDlg", w_top, 1);
    if (!psh_params) return;     /* create a property sheet */
    if ((psh_addline(psh_params, "lrange")              != 0)
    ||  (psh_additem(psh_params, "range",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "llrate")              != 0)
    ||  (psh_additem(psh_params, "lrate",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "llrdec")              != 0)
    ||  (psh_additem(psh_params, "lrdec",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "lnbhood")             != 0)
    ||  (psh_additem(psh_params, "nbhood", PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_params, "nbhood", "rect")      != 0)
    ||  (psh_menuadd(psh_params, "nbhood", "gauss")     != 0)
    ||  (psh_addline(psh_params, "lradius")             != 0)
    ||  (psh_additem(psh_params, "radius", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "lradec")              != 0)
    ||  (psh_additem(psh_params, "radec",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "lpatcnt")             != 0)
    ||  (psh_additem(psh_params, "patcnt", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "ldelay")              != 0)
    ||  (psh_additem(psh_params, "delay",  PSH_EDIT, 0) != 0)) {
      psh_delete(psh_params); error(E_DIALOG); return; }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%g", nnpar.range);      /* range of weights */
  psh_setval(psh_params, "range",  buf);
  sprintf(buf, "%g", nnpar.lrate);      /* learning rate */
  psh_setval(psh_params, "lrate",  buf);
  sprintf(buf, "%g", nnpar.lrdec);      /* learning rate decay par. */
  psh_setval(psh_params, "lrdec",  buf);
  psh_setval(psh_params, "nbhood",      /* neighborhood type */
            (char*)(ptrdiff_t)((nnpar.nbhood == SOM_RECT) ? 0 : 1));
  sprintf(buf, "%g", nnpar.radius);     /* neighborhood radius */
  psh_setval(psh_params, "radius", buf);
  sprintf(buf, "%g", nnpar.radec);      /* radius decay parameter */
  psh_setval(psh_params, "radec", buf);
  sprintf(buf, "%d", nnpar.patcnt);     /* patterns between redraws */
  psh_setval(psh_params, "patcnt", buf);
  sprintf(buf, "%g", nnpar.delay);      /* delay between redraws */
  psh_setval(psh_params, "delay",  buf);
  psh_handle(psh_params, dcb_params);   /* handle the dialog */
}  /* params() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  Cardinal n;                   /* argument counter */
  Colormap colmap;              /* color map of display */
  XColor   xc[9];               /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XSOM",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,           0) != 0)
  ||  (mn_additem (menu, "load",     1, mcb_load,    0) != 0)
  ||  (mn_additem (menu, "save",     0, mcb_save,    0) != 0)
  ||  (mn_additem (menu, "mpost",    0, mcb_mpost,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,           0) != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,    0) != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,           0) != 0)
  ||  (mn_additem (menu, "init",     0, mcb_init,    0) != 0)
  ||  (mn_additem (menu, "start",    2, mcb_start,   0) != 0)
  ||  (mn_additem (menu, "stop",     4, mcb_start,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,           0) != 0)
  ||  (mn_additem (menu, "view",     0, mcb_view,    0) != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw,  0) != 0)
  ||  (mn_addtitle(menu, "settings", 0, 0,           0) != 0)
  ||  (mn_additem (menu, "grid",     1, mcb_grid,    0) != 0)
  ||  (mn_additem (menu, "params",   0, mcb_params,  0) != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,           0) != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,   0) != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,           0) != 0))
    error(E_WIDGET);
  mn_resize (menu, INITXEXT);   /* set width of menu bar */
  mn_visible(menu, MN_ITEMS, 3);/* show/hide      menu items */
  mn_enable (menu, MN_ITEMS, 3);/* enable/disable menu items */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,     INITXEXT,
                        XtNheight,    INITYEXT, NULL);
  XtVaSetValues(w_top,  XtNminWidth,  MINXEXT,
                        XtNminHeight, MINYEXT,  NULL);
                                /* set window size */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */

  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  black  = BlackPixel(display, screen);
  white  = WhitePixel(display, screen);
  if (XAllocNamedColor(display, colmap, "grey", xc+1, xc) == 0)
    error(E_COLOR);             /* allocate colors and */
  grey   = xc[1].pixel;         /* note color pixel values */

  rseed((unsigned)time(NULL));  /* init. random number generator */
  proj3d = p3d_create();        /* create a projection object */
  if (!proj3d) error(E_NOMEM);  /* for sphere visualization */
  view(NULL,NULL,NULL,NULL);    /* and reset the view */

  /* --- evaluate command line arguments --- */
  if (argc > 1) dcb_load(NULL, (XtPointer)1, (XtPointer)argv[1]);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
